#include <stdio.h>
#include <orb/orbit.h>
#include "DOM.h"
#include "gdome.h"
#include "gnome-xml/tree.h"
#include "gnome-xml/parser.h"
#include "gdome-xml.h"

PortableServer_POA poa;

DOM_Element root;

GHashTable *nodemap = NULL;

typedef struct _GdomeElementG2C {
  POA_DOM_Element servant;
  GdomeElement *el;
} GdomeElementG2C;

extern POA_DOM_Element__vepv poa_dom_element_vepv;

static DOM_Element
gdome_element_g2c (GdomeElement *el)
{
  gpointer corba_object;

  if (nodemap == NULL)
    nodemap = g_hash_table_new (g_direct_hash, g_direct_equal);

  corba_object = g_hash_table_lookup (nodemap, el);
  if (corba_object == NULL)
    {
      /* create new corba object, register it in nodemap */
      GdomeElementG2C *servant;
      CORBA_Environment ev;

      CORBA_exception_init (&ev);

      servant = g_new0 (GdomeElementG2C, 1);
      servant->servant.vepv = &poa_dom_element_vepv;
      servant->el = el;

      POA_DOM_Element__init ((PortableServer_Servant) servant, &ev);

      PortableServer_POA_activate_object
	(poa, servant, &ev);

      corba_object = PortableServer_POA_servant_to_reference
	(poa, servant, &ev);

      g_hash_table_insert (nodemap, el, corba_object);
    }
  return corba_object;


}

static DOMString *
DOMString_make (const char *utf8)
{
  int byte_idx, short_idx;
  char c;
  int utf16len;
  int code;
  CORBA_unsigned_short *buf;
  DOMString *str;

  utf16len = 0;
  for (byte_idx = 0; (c = utf8[byte_idx]) != '\0'; byte_idx++)
    if (c & 0x80)
      {
	if (c & 0x40)
	  {
	    if ((c & 0xf8) != 0xf0)
	      utf16len += 1;
	    else
	      utf16len += 2;
	  }
      }
    else
      utf16len++;

  buf = CORBA_sequence_CORBA_unsigned_short_allocbuf (utf16len);

  short_idx = 0;
  for (byte_idx = 0; (c = utf8[byte_idx]) != '\0';)
    {
      if (c & 0x80)
	{
	  if ((c & 0xe0) == 0xe0)
	    {
	      if ((c & 0xf0) == 0xf0)
		{
		  /* 4-byte code */
		  code = ((c & 0x07) << 18) |
		    ((utf8[byte_idx + 1] & 0x3f) << 12) |
		    ((utf8[byte_idx + 2] & 0x3f) << 6) |
		    (utf8[byte_idx + 3] & 0x3f);
		  /* an alternative "raphcode" sequence would be
		     ((c << 18) | (utf8[byte_idx + 1] << 12) |
		     (utf8[byte_idx + 2] << 6) | utf8[byte_idx + 3) -
		     0x3c82080 */
		  byte_idx += 4;
		}
	      else
		{
		  /* 3-byte code */
		  code = ((c & 0x0f) << 12) |
		    ((utf8[byte_idx + 1] & 0x3f) << 6) |
		    (utf8[byte_idx + 2] & 0x3f);
		  /* alternately,
		     ((c << 12) | (utf8[byte_idx + 1] << 6) |
		     utf8[byte_idx + 2]) - 0xe2080 */
		  byte_idx += 3;
		}
	    }
	  else
	    {
	      /* 2-byte code */
	      code = ((c & 0x1f) << 6) |
		(utf8[byte_idx + 1] & 0x3f);
	      /* alternately,
		 ((c << 6) | utf8[byte_idx + 1]) - 0x3080 */
	      byte_idx += 2;
	    }
	}
      else
	{
	  /* ASCII */
	  code = c;
	  byte_idx++;
	}
      if (code > 0xffff)
	{
	  buf[short_idx++] = 0xd800 | ((code - 0x10000) >> 10);
	  buf[short_idx++] = 0xdc00 | (code & 0x3fff);
	}
      else
	buf[short_idx++] = code;
    }

  str = DOMString__alloc();
  str->_maximum = 0;
  str->_length = short_idx;
  str->_buffer = buf;
  CORBA_sequence_set_release(str, CORBA_TRUE);

  return str;
}

static DOMString *
DOMString_g2c (GdomeDOMString *gstr)
{
  DOMString *cstr;

  cstr = DOMString_make (gstr->str);
  gdome_str_unref (gstr);
  return cstr;
}

static DOMString *
element__get_tagName (PortableServer_Servant servant,
			       CORBA_Environment *ev)
{
  GdomeElementG2C *g2c_el = (GdomeElementG2C *)servant;
  return DOMString_g2c (gdome_el_tagName(g2c_el->el));
}

static CORBA_Object
document__get_documentElement (PortableServer_Servant servant,
			       CORBA_Environment *ev)
{
  return CORBA_Object_duplicate (root, ev);
}

PortableServer_ServantBase__epv base_epv = {
  NULL,
  NULL,
  NULL
};
POA_DOM_Node__epv node_epv = { NULL };
POA_DOM_Element__epv element_epv = { NULL,
				      element__get_tagName };
POA_DOM_Document__epv document_epv = { NULL,
				       NULL,
				       NULL,
				       document__get_documentElement };
POA_DOM_Element__vepv poa_dom_element_vepv = { &base_epv, &node_epv, &element_epv };
POA_DOM_Document__vepv poa_dom_document_vepv = { &base_epv, &node_epv, &document_epv };
POA_DOM_Element poa_dom_element_servant = { NULL, &poa_dom_element_vepv };
POA_DOM_Document poa_dom_document_servant = { NULL, &poa_dom_document_vepv };

DOM_Document gdome_client = CORBA_OBJECT_NIL;

int
main (int argc, char **argv)
{
  PortableServer_ObjectId objid = {0, sizeof("myGdomeServer"), "myGdomeServer"};
  PortableServer_ObjectId objid2 = {0, sizeof("myGdomeServer2"), "myGdomeServer2"};

  CORBA_Environment ev;
  char *retval;
  CORBA_ORB orb;

  xmlDocPtr doc;
  GdomeDocument *gdoc;

  CORBA_exception_init (&ev);
  orb = CORBA_ORB_init (&argc, argv, "orbit-local-orb", &ev);

  POA_DOM_Document__init (&poa_dom_document_servant, &ev);
  POA_DOM_Element__init (&poa_dom_element_servant, &ev);

  poa = (PortableServer_POA)CORBA_ORB_resolve_initial_references(orb, "RootPOA", &ev);
  PortableServer_POAManager_activate(PortableServer_POA__get_the_POAManager(poa, &ev), &ev);
  PortableServer_POA_activate_object_with_id(poa,
					     &objid2, &poa_dom_element_servant, &ev);
  PortableServer_POA_activate_object_with_id(poa,
					     &objid, &poa_dom_document_servant, &ev);

  gdome_client = PortableServer_POA_servant_to_reference(poa,
							  &poa_dom_document_servant,
							  &ev);

  root = PortableServer_POA_servant_to_reference(poa,
							  &poa_dom_element_servant,
							  &ev);


  if (argc < 2)
    {
      fprintf (stderr, "usage: gdome-server test.xml\n");
      exit (1);
    }

  doc = xmlParseFile (argv[1]);
  gdoc = gdome_xml_from_document (doc);
  root = gdome_element_g2c (gdome_doc_documentElement (gdoc));

  if (!gdome_client) {
    printf("Cannot get objref\n");
    return 1;
  }

  retval = CORBA_ORB_object_to_string(orb, gdome_client, &ev);

  g_print("%s\n", retval); fflush(stdout);

  CORBA_free(retval);

  CORBA_ORB_run(orb, &ev);

  return 0;
}
